"""
    omniCLIP is a CLIP-Seq peak caller

    Copyright (C) 2017 Philipp Boss

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import gffutils

# Gene ID identifiers
id_spec = {"gene": ["gene_id", "geneID", "ID=gene:"]}


def CreateDB(FileNameGFF, FileNameDB, gene_features):
    """Create a GFF database using GFFUtils."""
    # Read the annotation file
    with open(FileNameGFF, "r") as f:
        lines = f.readlines()

    # Keep only lines describing genes as features
    genes = "\n".join(
        [
            "\t".join(line.split("\t")[:2] + ["gene"] + line.split("\t")[3:])
            for line in lines
            if (line[0] != "#" and line.split("\t")[2] in gene_features)
        ]
    )

    # Create the DB
    gffutils.create_db(
        data=genes,
        from_string=True,
        dbfn=FileNameDB,
        force=True,
        keep_order=True,
        merge_strategy="create_unique",
        disable_infer_transcripts=True,
        disable_infer_genes=True,
        sort_attribute_values=True,
        id_spec=id_spec,
    )



from distutils.command.build_ext import build_ext

import numpy as np
from Cython.Build import cythonize
from setuptools import Extension
from setuptools.dist import Distribution


def build(setup_kwargs):
    """Needed for the poetry building interface."""
    extensions = [
        Extension(
            "omniCLIP.viterbi",
            sources=["omniCLIP/omni_stat/viterbi.pyx"],
            include_dirs=[np.get_include()],
        ),
    ]
    extensions = cythonize(
        extensions,
        compiler_directives={"language_level": 3, "linetrace": True},
    )
    setup_kwargs.update(
        {
            "ext_modules": extensions,
            "cmdclass": {"build_ext": build_ext},
            "include_dirs": [np.get_include()],
            "zip_safe": False,
        }
    )
    return setup_kwargs



"""
    omniCLIP is a CLIP-Seq peak caller

    Copyright (C) 2017 Philipp Boss

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from collections import defaultdict
from intervaltree import IntervalTree
import numpy as np
from scipy.sparse import csr_matrix


def mask_miRNA_positions(Sequences, GeneAnnotation):
    """Remove counts overlapping miRNA positions.

    This function takes the sequences and the gene annotation and sets all
    counts in Sequences to zero that overlap miRNAs in the gene annotation.
    """
    keys = ["Coverage", "Read-ends", "Variants"]
    # Create a dictionary that stores the genes in the Gene annnotation
    gene_dict = {}

    for gene in GeneAnnotation.features_of_type("gene"):
        gene_dict[gene.id] = gene

    # Get Chromosomes:
    genes_chr_dict = defaultdict(list)
    for gene in list(gene_dict.values()):
        genes_chr_dict[gene.chrom].append(gene)

    # Create an interval tree for the genes:
    interval_chr_dict = {}
    for chr in list(genes_chr_dict.keys()):
        interval_chr_dict[chr] = IntervalTree()
        for gene in genes_chr_dict[chr]:
            if gene in list(Sequences.keys()):
                interval_chr_dict[chr][gene.start : gene.stop] = gene

    # Iterate over the genes in the Sequences:
    miRNAs = [
        miRNA
        for miRNA in GeneAnnotation.features_of_type("gene")
        if miRNA.attributes.get("gene_type").count("miRNA") > 0
    ]
    for miRNA in miRNAs:
        curr_chr = miRNA.chrom
        curr_genes = [
            gene[2]
            for gene in sorted(interval_chr_dict[curr_chr][miRNA.start : miRNA.stop])
        ]

        # Get the miRNAs that overlap:
        for curr_gene_obj in curr_genes:
            curr_gene = curr_gene_obj.id

            # Get position relative to the host gene
            curr_start = max(0, miRNA.start - gene_dict[curr_gene].start)
            curr_stop = max(
                gene_dict[curr_gene].stop - gene_dict[curr_gene].start,
                miRNA.stop - gene_dict[curr_gene].start,
            )

            # Set for each field the sequences to zeros
            for curr_key in keys:
                if curr_key in Sequences[curr_gene]:
                    for seq in list(Sequences[curr_gene][curr_key].values()):
                        if curr_key == "Variants":
                            # Convert the Variants to array
                            curr_seq = csr_matrix(
                                (seq["data"][:], seq["indices"][:], seq["indptr"][:]),
                                shape=seq["shape"][:],
                            )

                            ix_slice = np.logical_and(
                                curr_start <= curr_seq.indices,
                                curr_seq.indices < curr_stop,
                            )

                            seq["data"][ix_slice] = 0
                        else:
                            seq[:, curr_start:curr_stop] = 0


def mask_overlapping_positions(Sequences, GeneAnnotation):
    """Generate an overlapping regions track."""
    # Add fields to Sequence structure:
    for gene in list(Sequences.keys()):
        Sequences[gene].create_group("mask")
        rep = list(Sequences[gene]["Coverage"].keys())[0]
        if rep == "0":
            Sequences[gene]["mask"].create_dataset(
                rep,
                data=np.zeros(Sequences[gene]["Coverage"][rep][()].shape),
                compression="gzip",
                compression_opts=9,
                chunks=Sequences[gene]["Coverage"][rep][()].shape,
                dtype="i8",
            )

    # Create a dictionary that stores the genes in the Gene annnotation
    genes = [gene for gene in GeneAnnotation.features_of_type("gene")]
    gene_dict = {gene.id: gene for gene in genes}

    # Get Chromosomes:
    genes_chr_dict = defaultdict(list)
    for gene in list(gene_dict.values()):
        genes_chr_dict[gene.chrom].append(gene)

    # Create an interval tree for the genes:
    interval_chr_dict = {}
    for chr in list(genes_chr_dict.keys()):
        interval_chr_dict[chr] = IntervalTree()
        for gene in genes_chr_dict[chr]:
            interval_chr_dict[chr][gene.start : gene.stop] = gene

    # Iterate over the genes in the Sequences:
    for gene in genes:
        if gene.id in Sequences:
            curr_chr = gene.chrom
            curr_genes = sorted(interval_chr_dict[curr_chr][gene.start : gene.stop])
            curr_genes = [curr_gene[2] for curr_gene in curr_genes]
            curr_genes.remove(gene)
            # Get the genes that overalp:
            for curr_gene_obj in curr_genes:
                curr_gene = curr_gene_obj.id

                # Get position of overlapping gene relative to the host gene
                ovrlp_start = max(0, gene_dict[curr_gene].start - gene.start)
                ovrlp_stop = min(
                    gene.stop - gene.start, gene_dict[curr_gene].stop - gene.start
                )

                # Set for each field the sequences to zeros
                rep = list(Sequences[gene.id]["Coverage"].keys())[0]
                Sequences[gene.id]["mask"][rep][0, ovrlp_start:ovrlp_stop] = True



"""
    omniCLIP is a CLIP-Seq peak caller

    Copyright (C) 2017 Philipp Boss

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
from scipy.special import psi

from . import multdirichletVect


def ComputeStateProbForGeneMD_unif(*args):
    """Compute the prior for the dirichlet model."""
    Counts, alpha, State, EmissionParameters = args

    tracks_per_rep = alpha.shape[0]

    NrOfReplicates = Counts.shape[0] // tracks_per_rep

    Prob = np.zeros((Counts.shape[1]))

    # Precompute the Zero values
    IxZeros = np.sum(Counts, axis=0) == 0
    IxNonZeros = np.sum(Counts, axis=0) > 0

    # If there are zero values, add them to Counts in the first column
    if np.sum(IxZeros) > 0:
        ZeroCounts = np.zeros((Counts.shape[0], 1))
        Counts = np.hstack((ZeroCounts, Counts[:, IxNonZeros]))
    else:
        Counts = Counts[:, IxNonZeros]

    RatioLikelihood = multdirichletVect.log_pdf_vect(Counts[0:tracks_per_rep, :], alpha)
    for i in range(1, NrOfReplicates):
        RatioLikelihood += multdirichletVect.log_pdf_vect(
            Counts[i * tracks_per_rep : (i + 1) * tracks_per_rep, :], alpha
        )

    # Compute the likelihood for the current observation for the current model models
    if np.sum(IxZeros) > 0:
        Prob[IxNonZeros] = RatioLikelihood[0, 1:]
        Prob[IxZeros] = np.tile(RatioLikelihood[0, 0], (1, np.sum(IxZeros)))
    else:
        Prob = RatioLikelihood[0, :]

    return Prob


def ComputeStateProbForGeneMD_unif_rep(*args):
    """Compute the prior for the dirichlet model."""
    Counts, alpha, State, EmissionParameters = args

    tracks_per_rep = alpha.shape[0]
    NrOfReplicates = Counts.shape[0] // tracks_per_rep

    Prob = np.zeros((Counts.shape[1]))

    # Precompute the Zero values
    IxZeros = np.sum(Counts, axis=0) == 0
    IxNonZeros = np.sum(Counts, axis=0) > 0

    # If there are zero values, add them to Counts in the first column
    if np.sum(IxZeros) > 0:
        ZeroCounts = np.zeros((Counts.shape[0], 1))
        Counts = np.hstack((ZeroCounts, Counts[:, IxNonZeros]))
    else:
        Counts = Counts[:, IxNonZeros]

    RatioLikelihood = multdirichletVect.log_pdf_vect_rep(
        Counts, alpha, tracks_per_rep, NrOfReplicates
    )
    # Compute the likelihood for the current observation for the current model models

    if np.sum(IxZeros) > 0:
        Prob[IxNonZeros] = RatioLikelihood[0, 1:]
        Prob[IxZeros] = RatioLikelihood[0, 0]
    else:
        Prob = RatioLikelihood[0, :]

    return Prob


def MDK_f_joint_vect_unif(x, *args):
    """Compute the likelihood of the parameters."""
    alpha = x
    Counts, NrOfCounts, EmissionParameters = args
    # Prepare the return variable
    LogLikelihood = 0.0
    NrOfReplicates = Counts.shape[0] // alpha.shape[0]
    tracks_per_rep = x.shape[0]

    RatioLikelihood = multdirichletVect.log_pdf_vect_rep(
        Counts, alpha, tracks_per_rep, NrOfReplicates
    )
    CurrLogLikelihood = RatioLikelihood * NrOfCounts
    LogLikelihood = np.sum(CurrLogLikelihood[np.isinf(CurrLogLikelihood) == 0])

    return -LogLikelihood


def MD_f_joint_vect_unif(x, *args):
    """Compute the likelihood of the parameters."""
    alpha = x
    Counts, NrOfCounts, EmissionParameters = args

    # Prepare the return variable
    LogLikelihood = 0.0

    tracks_per_rep = x.shape[0]
    NrOfReplicates = Counts.shape[0] // tracks_per_rep

    RatioLikelihood = multdirichletVect.log_pdf_vect(Counts[0:tracks_per_rep, :], alpha)
    for i in range(1, NrOfReplicates):
        RatioLikelihood += multdirichletVect.log_pdf_vect(
            Counts[i * tracks_per_rep : (i + 1) * tracks_per_rep, :], alpha
        )
    CurrLogLikelihood = RatioLikelihood * np.float64(NrOfCounts)
    LogLikelihood += np.sum(CurrLogLikelihood[np.isinf(CurrLogLikelihood) == 0])

    return -LogLikelihood


def MD_f_prime_joint_vect_unif(x, *args):
    Counts, NrOfCounts, EmissionParameters = args

    tracks_per_rep = x.shape[0]
    NrOfReplicates = Counts.shape[0] // tracks_per_rep

    # Prepare the return variable
    LogLikelihood = np.zeros_like(x, dtype=np.float)

    # Compute the likelihood
    curr_k = Counts[0:tracks_per_rep, :]
    curr_alpha = x

    DBase = NrOfReplicates * psi(np.sum(curr_alpha)) - psi(
        np.sum(curr_k, axis=0) + np.sum(curr_alpha)
    )
    for rep in range(1, NrOfReplicates):
        curr_k = Counts[rep * tracks_per_rep : (rep + 1) * tracks_per_rep, :]
        DBase -= psi(np.sum(curr_k, axis=0) + np.sum(curr_alpha))

    for J in range(0, curr_alpha.shape[0]):
        curr_k = Counts[J, :]
        ix_zero = curr_k == 0

        D = np.float64(np.zeros_like(curr_k))
        if np.isscalar(D):
            D = psi(curr_alpha[J] + curr_k) - psi(curr_alpha[J])
        else:
            D[ix_zero] = psi(curr_alpha[J] + curr_k[ix_zero]) - psi(curr_alpha[J])

        for rep in range(1, NrOfReplicates):
            curr_k = Counts[rep * tracks_per_rep + J, :]
            ix_zero = curr_k == 0
            if np.isscalar(D):
                D += psi(curr_alpha[J] + curr_k) - psi(curr_alpha[J])
            else:
                D[ix_zero] += psi(curr_alpha[J] + curr_k[ix_zero]) - psi(curr_alpha[J])

        CurrLogLikeliehood = np.float64((D + DBase) * np.float64(NrOfCounts))
        LogLikelihood[J] += np.sum(
            CurrLogLikeliehood[np.isinf(CurrLogLikeliehood) == 0]
        )
    return -LogLikelihood



"""omniCLIP is a CLIP-Seq peak caller.

Copyright (C) 2017 Philipp Boss

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import random
import shutil

import numpy as np

from . import LoadReads, tools


def parsing_argparse(args):
    """Parse the argparse dict."""
    # Listing args to extract from the argparse parser
    named_args = [
        "bg_type",
        "dat_file_bg",
        "dat_file_clip",
        "diag_bg",
        "diag_event_mod",
        "emp_var",
        "fg_pen",
        "filter_snps",
        "glm_weight",
        "ign_diag",
        "ign_GLM",
        "ign_out_rds",
        "mask_ovrlp",
        "max_it",
        "nb_proc",
        "norm_class",
        "out_dir",
        "pv_cutoff",
        "skip_diag_event_mdl",
        "snps_min_cov",
        "snps_thresh",
        "subs",
        "tmp_dir",
        "verbosity",
    ]

    # Creating the params dictionary
    params = {arg: getattr(args, arg) for arg in named_args}

    # Verifying the validity of the args
    # verifying_args(params)

    # Printing args if verbosity
    if params["verbosity"] > 1:
        print(args)

    # Defining fixed-value params
    params["ExpressionParameters"] = [None, None]
    params["LastIter"] = False

    # Conditional params - Number of States
    if params["bg_type"] == "Coverage_bck":
        params["NrOfStates"] = 4
    else:
        params["NrOfStates"] = 3

    # Conditional params - Ignore reads outside gene and diag model
    if params["ign_out_rds"]:
        params["ign_diag"] = params["ign_out_rds"]

    # Conditional params - Prior Matrix
    params["PriorMatrix"] = np.ones((params["NrOfStates"], 1)) / float(
        params["NrOfStates"]
    )

    # Conditional  params - Transition Matrix
    TransMat = np.ones((params["NrOfStates"], params["NrOfStates"])) + np.eye(
        params["NrOfStates"]
    )
    params["TransMat"] = TransMat / np.sum(np.sum(TransMat))

    # Conditional params - Diag Event Params, flag for penalized path pred
    params["Diag_event_params"] = {}
    params["Diag_event_params"]["nr_mix_comp"] = args.nr_mix_comp
    params["Diag_event_params"]["alpha"] = {}
    params["Diag_event_params"]["mix_comp"] = {}
    for state in range(params["NrOfStates"]):
        mixtures = np.random.uniform(0.0, 1.0, size=(args.nr_mix_comp))
        params["Diag_event_params"]["mix_comp"][state] = mixtures / np.sum(
            mixtures
        )

    # Conditional params - out file name
    params["out_file_base"] = "pred"
    if params["ign_GLM"]:
        params["out_file_base"] += "_no_glm"
    if params["ign_diag"]:
        params["out_file_base"] += "_no_diag"

    # Conditional params - tmp_dir
    # If no tmp_dir is specified, will place files in the CLIP files location
    if params["tmp_dir"] is None:
        params["tmp_dir"] = os.path.dirname(
            os.path.realpath(params["dat_file_clip"])
        )

    # Optional params - Random seeding
    if args.rnd_seed is not None:
        random.seed(args.rnd_seed)
        print("setting random seed")

    return params


def dup_seqfiles(params):
    """Create temporary Seqfiles that will be modified by omniCLIP."""
    fg_tmp = os.path.join(params["tmp_dir"], "clip_data.dat.tmp")
    bg_tmp = os.path.join(params["tmp_dir"], "bg_data.dat.tmp")

    os.makedirs(os.path.dirname(fg_tmp), exist_ok=True)
    shutil.copy(params["dat_file_clip"], fg_tmp)
    shutil.copy(params["dat_file_bg"], bg_tmp)

    # Open the temporary read files
    params["dat_file_clip"] = fg_tmp
    params["dat_file_bg"] = bg_tmp

    return params


def parsing_files(args, params):
    """Parse arguments that are function of the data files."""
    # Loading Sequence and Background
    Sequences = LoadReads.get_data_handle(params["dat_file_clip"], write=True)
    Background = LoadReads.get_data_handle(params["dat_file_bg"], write=True)

    # Estimate the library size
    params["LibrarySize"] = tools.estimate_library_size(Sequences)
    params["BckLibrarySize"] = tools.estimate_library_size(Background)

    # Estimate to number of replicates
    gene = list(Sequences.keys())[0]
    params["NrOfReplicates"] = len(Sequences[gene]["Coverage"])
    params["NrOfBckReplicates"] = len(Background[gene]["Coverage"])

    # Initialise the parameter vector alpha
    alphashape = (
        Sequences[gene]["Variants"]["0"]["shape"][0]
        + Sequences[gene]["Coverage"]["0"][()].shape[0]
        + Sequences[gene]["Read-ends"]["0"][()].shape[0]
    )
    for state in range(params["NrOfStates"]):
        params["Diag_event_params"]["alpha"][state] = np.random.uniform(
            0.9, 1.1, size=(alphashape, args.nr_mix_comp)
        )

    return params



"""
    omniCLIP is a CLIP-Seq peak caller

    Copyright (C) 2017 Philipp Boss

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from collections import defaultdict
from scipy.special import logsumexp
from scipy.sparse import csr_matrix
from scipy.stats import nbinom
import h5py
import itertools
import multiprocessing
import numpy as np
import pandas as pd
import time

from . import LoadReads
from .. import viterbi
from ..omni_stat import diag_event_model
from ..omni_stat import emission_prob
from ..omni_stat import trans


def GetModelIx(
    Sequences, Type="all", snps_thresh=0.4, snps_min_cov=10, Background=None
):
    """Return the positions for at which the emission should be computed."""
    if Type == "all":
        Ix = np.sum(StackData(Sequences, add="all"), axis=0) > 0
    elif Type == "no_snps_all":
        if Background is not None:
            Bck_var = np.sum(StackData(Background, add="only_var"), axis=0)
            Bck = np.sum(StackData(Background, add="only_cov"), axis=0)
            Bck_ratio = np.zeros_like(Bck_var)
            ix = Bck > 0
            Bck_ratio[ix] = np.float64(Bck_var[ix]) / np.float64(Bck[ix])
            # Positions that have a minimal coverage and have a snp ratio above
            # the threshold are not considered
            Ix_bg = (Bck > snps_min_cov) * (Bck_ratio > snps_thresh) > 0
            Ix_fg = np.sum(StackData(Sequences, add="all"), axis=0) > 0
            Ix = (Ix_fg > 0) * (Ix_bg == 0) > 0
    elif Type == "no_snps_conv":
        if Background is not None:
            Bck_var = np.sum(StackData(Background, add="only_var"), axis=0)
            Bck = np.sum(StackData(Background, add="only_cov"), axis=0)
            Bck_ratio = np.zeros_like(Bck_var)
            ix = Bck > 0
            Bck_ratio[ix] = np.float64(Bck_var[ix]) / np.float64(Bck[ix])
            # Positions that have a minimal coverage and have a snp ratio above
            # the threshold are not considered
            Ix_bg = (Bck > snps_min_cov) * (Bck_ratio > snps_thresh) > 0
            Ix_fg = np.sum(StackData(Sequences, add="nocov"), axis=0) > 0
            Ix = (Ix_fg > 0) * (Ix_bg == 0) > 0
    elif Type == "Conv":
        Ix = np.sum(StackData(Sequences, add="nocov"), axis=0) > 0

    else:
        Ix = np.sum(StackData(Sequences, add="nocov"), axis=0) > 0
    return Ix


def StackData(Sequences, add="all", use_strand="True"):
    """Stack the data for a gene."""
    Track_strand_map = [
        18,
        17,
        16,
        15,
        19,
        13,
        12,
        11,
        10,
        14,
        8,
        7,
        6,
        5,
        9,
        3,
        2,
        1,
        0,
        4,
    ]

    repls = list(Sequences["Coverage"].keys())
    nr_of_repl = len(repls)
    gene_len = Sequences["Coverage"]["0"].shape[1]
    if (add == "all") or (add == "nocov"):
        CurrStack = np.zeros((nr_of_repl, gene_len))
        for rep in repls:
            CurrStack[int(rep), :] = Sequences["SummedCoverage"][rep]

    elif add == "only_cov":
        # Check if the cariants are substracted from the coverage:
        if "Variants" in Sequences:
            CurrStack = np.zeros((nr_of_repl, gene_len))
            for rep in repls:
                CurrStack[int(rep), :] = Sequences["SummedCoverage"][rep]
        else:
            CurrStack = np.zeros((nr_of_repl, gene_len))
            for rep in repls:
                CurrStack[int(rep), :] += Sequences["Coverage"][rep].sum(axis=0)
    elif add == "only_var":
        CurrStack = np.zeros((nr_of_repl, gene_len))
        for rep in repls:
            CurrStack[int(rep), :] += Sequences["Variants"][rep].sum(axis=0)

    elif add == "only_var_summed":
        CurrStack = np.zeros((Sequences["Variants"][repls[0]].shape[0], gene_len))
        for rep in repls:
            if use_strand and (not (Sequences["strand"] == 1)):
                CurrStack += Sequences["Variants"][rep][Track_strand_map, :]
            else:
                CurrStack += Sequences["Variants"][rep]

    else:
        # First compute dimensions of return array:
        nr_rows = 0
        for rep in list(Sequences["Variants"].keys()):
            nr_cols = Sequences["Coverage"][rep].shape[1]
            nr_rows += (
                Sequences["Variants"][rep].shape[0]
                + Sequences["Read-ends"][rep].shape[0]
                + Sequences["Coverage"][rep].shape[0]
            )

        CurrStack = np.zeros((nr_rows, nr_cols))
        row_counter = 0
        for rep in list(Sequences["Variants"].keys()):
            if use_strand and (not (Sequences["strand"] == 1)):
                CurrStack[
                    row_counter : (row_counter + Sequences["Variants"][rep].shape[0]), :
                ] = Sequences["Variants"][rep][Track_strand_map, :]
            else:
                CurrStack[
                    row_counter : (row_counter + Sequences["Variants"][rep].shape[0]), :
                ] = Sequences["Variants"][rep]
            row_counter += Sequences["Variants"][rep].shape[0]

            CurrStack[
                row_counter : (row_counter + Sequences["Read-ends"][rep].shape[0]), :
            ] = Sequences["Read-ends"][rep]
            row_counter += Sequences["Read-ends"][rep].shape[0]

            CurrStack[
                row_counter : (row_counter + Sequences["Coverage"][rep].shape[0]), :
            ] = Sequences["Coverage"][rep]
            row_counter += Sequences["Coverage"][rep].shape[0]

    CurrStack[CurrStack < 0] = 0
    return CurrStack


def PreloadSequencesForGene(Sequences, gene):
    """Stacks the data for a gene."""
    Sequences_per_gene = {}
    for key in Sequences[gene]:
        if isinstance(Sequences[gene][key], h5py.Dataset):
            Sequences_per_gene[key] = Sequences[gene][key][()]
        else:
            Sequences_per_gene[key] = {}
            for rep in Sequences[gene][key]:
                if key == "Variants":
                    # Convert the Variants to array again
                    Sequences_per_gene[key][rep] = csr_matrix(
                        (
                            Sequences[gene]["Variants"][rep]["data"][:],
                            Sequences[gene]["Variants"][rep]["indices"][:],
                            Sequences[gene]["Variants"][rep]["indptr"][:],
                        ),
                        shape=Sequences[gene]["Variants"][rep]["shape"][:],
                    ).toarray()
                else:
                    Sequences_per_gene[key][rep] = Sequences[gene][key][rep][()]

    return Sequences_per_gene


def GetSuffStat(
    Paths,
    NrOfStates,
    Type,
    ResetNotUsedStates=True,
    EmissionParameters=None,
    verbosity=1,
):
    """Compute for each CurrPath state a set of sufficient statistics."""
    # Initialize the sufficent statistics variable
    print("Getting suffcient statistic")
    t = time.time()
    SuffStat = {}
    for CurrState in range(NrOfStates):
        SuffStat[CurrState] = defaultdict(int)

    Sequences = h5py.File(EmissionParameters["dat_file_clip"], "r")
    Background = h5py.File(EmissionParameters["dat_file_bg"], "r")

    # Fill the sufficent statistics variable
    for gene in list(Sequences.keys()):
        rep = list(Sequences[gene]["Coverage"].keys())[0]
        CurrGenePath = Paths[gene]

        # Stack the matrices together and convert to dense matrix
        Sequences_per_gene = PreloadSequencesForGene(Sequences, gene)
        Background_per_gene = PreloadSequencesForGene(Background, gene)
        if Type == "Conv":
            CurrStack = StackData(Sequences_per_gene, add="variants")
        else:
            CurrStack = StackData(Sequences_per_gene, add="all")

        if EmissionParameters["filter_snps"]:
            if Type == "Conv":
                Ix = GetModelIx(
                    Sequences_per_gene,
                    Type="no_snps_conv",
                    snps_thresh=EmissionParameters["snps_thresh"],
                    snps_min_cov=EmissionParameters["snps_min_cov"],
                    Background=Background_per_gene,
                )
            else:
                Ix = GetModelIx(Sequences_per_gene, Type)
        else:
            Ix = GetModelIx(Sequences_per_gene, Type)

        NonZero = np.sum(CurrStack, axis=0) > 0

        # Determine the nonzeros elements
        for CurrState in range(NrOfStates):
            if EmissionParameters["mask_ovrlp"]:
                CurrIx = (
                    Ix
                    * (Sequences_per_gene["mask"][rep][0, :] == 0)
                    * NonZero
                    * (CurrGenePath == CurrState)
                    > 0
                )
            else:
                CurrIx = Ix * NonZero * (CurrGenePath == CurrState) > 0

            data = CurrStack[:, CurrIx].T
            ncols = data.shape[1]
            dtype = data.T.dtype.descr * ncols
            struct = data.view(dtype)

            vals, val_counts = np.unique(struct, return_counts=True)

            # Save the tuples and how many times they have been seen so far.
            for curr_val, curr_count in zip(vals, val_counts):
                SuffStat[CurrState][tuple(curr_val)] += curr_count

            # Treat the 0 tuple seperately for speed improvment
            if len(Ix) == 0:
                continue

            NullIx = (NonZero == 0) * (CurrGenePath == CurrState) > 0
            if np.sum(NullIx) == 0:
                continue
            NullCount = np.sum(NullIx)
            if NullCount > 0:
                NullTuple = np.zeros_like(CurrStack[:, 0])
                NullTuple = tuple(NullTuple.T)
                SuffStat[CurrState][NullTuple] += NullCount

        del CurrStack, NonZero, CurrGenePath, Ix
    if verbosity > 0:
        print("Done: Elapsed time: " + str(time.time() - t))

    return SuffStat


def GetSuffStatBck(
    Paths,
    NrOfStates,
    Type,
    ResetNotUsedStates=True,
    EmissionParameters=None,
    verbosity=1,
):
    """Computes for each CurrPath state a set of sufficient statistics."""
    # Initialize the sufficent statistics variable
    print("Getting suffcient statistic")
    t = time.time()
    SuffStatBck = {}

    fg_state, bg_state = emission_prob.get_fg_and_bck_state(
        EmissionParameters, final_pred=True
    )

    SuffStatBck[fg_state] = defaultdict(int)

    Sequences = h5py.File(EmissionParameters["dat_file_clip"], "r")
    Background = h5py.File(EmissionParameters["dat_file_bg"], "r")

    # Fill the sufficient statistics variable
    for gene in list(Sequences.keys()):
        rep = list(Background[gene]["Coverage"].keys())[0]
        CurrGenePath = Paths[gene]

        # Stack the matrices together and convert to dense matrix
        Background_per_gene = PreloadSequencesForGene(Background, gene)
        Sequences_per_gene = PreloadSequencesForGene(Sequences, gene)
        if Type == "Conv":
            CurrStack = StackData(Background_per_gene, add="variants")
        else:
            CurrStack = StackData(Background_per_gene, add="all")

        if EmissionParameters["filter_snps"]:
            if Type == "Conv":
                Ix = GetModelIx(
                    Background_per_gene,
                    Type="no_snps_conv",
                    snps_thresh=EmissionParameters["snps_thresh"],
                    snps_min_cov=EmissionParameters["snps_min_cov"],
                    Background=Background_per_gene,
                )
            else:
                Ix = GetModelIx(Background_per_gene, Type)
        else:
            Ix = GetModelIx(Background_per_gene, Type)

        NonZero = np.sum(CurrStack, axis=0) > 0

        # Determine the nonzeros elements
        CurrState = fg_state

        CurrIx = Ix * NonZero > 0
        if EmissionParameters["mask_ovrlp"]:
            CurrIx = (
                Ix
                * NonZero
                * (CurrGenePath == CurrState)
                * (Sequences_per_gene["mask"][rep][0, :] == 0)
            ) > 0
        else:
            CurrIx = Ix * NonZero * (CurrGenePath == CurrState) > 0

        data = CurrStack[:, CurrIx].T
        ncols = data.shape[1]
        dtype = data.T.dtype.descr * ncols
        struct = data.view(dtype)

        vals, val_counts = np.unique(struct, return_counts=True)

        # Save the tuples and how many times they have been seen so far.
        for curr_val, curr_count in zip(vals, val_counts):
            SuffStatBck[CurrState][tuple(curr_val)] += curr_count

        # Treat the 0 tuple seperately for speed improvment
        if len(Ix) == 0:
            continue
        NullIx = (NonZero == 0) * (CurrGenePath == CurrState) > 0
        if np.sum(NullIx) == 0:
            continue
        NullCount = np.sum(NullIx)
        if NullCount > 0:
            NullTuple = np.zeros_like(CurrStack[:, 0])
            NullTuple = tuple(NullTuple.T)
            SuffStatBck[CurrState][NullTuple] += NullCount

        del CurrStack, NonZero, CurrGenePath, Ix

    print("Done: Elapsed time: " + str(time.time() - t))

    return SuffStatBck


def ConvertSuffStatToArrays(SuffStat):
    """Convert the sufficient statistics into a list of arrays."""
    # Initialise the return values
    Counts = {}
    NrOfCounts = {}
    for State in list(SuffStat.keys()):
        if len(list(SuffStat[State].keys())) == 0:
            print("empyt suffcient keys ")
            Counts[State] = np.array([])
            NrOfCounts[State] = np.array([])
        else:
            Counts[State] = np.array(
                [np.array(key) for key in list(SuffStat[State].keys())]
            ).T
            NrOfCounts[State] = np.tile(
                np.array(list(SuffStat[State].values())), (1, 1)
            )

    return Counts, NrOfCounts


def repl_track_nr(ex_list, offset):
    """Compute replicate list offset.

    Compute for a list of tracks in one replicate additionaly the list for the
    second replicate.
    """
    return ex_list + list(np.array(ex_list) + offset)


def GeneratePred(
    Paths,
    Sequences,
    Background,
    IterParameters,
    GeneAnnotation,
    OutFile,
    fg_state=1,
    noise_state=0,
    pv_cutoff=0.05,
    verbosity=1,
):
    """Write the predictions."""
    TransitionParameters = IterParameters[1]
    EmissionParameters = IterParameters[0]

    merge_neighbouring_sites = False
    minimal_site_length = 1

    # Predict the sites
    print("Score peaks")
    LoadReads.close_data_handles(handles=[Sequences, Background])
    Sequences = h5py.File(EmissionParameters["dat_file_clip"], "r")
    Background = h5py.File(EmissionParameters["dat_file_bg"], "r")

    ScoredSites = GetSites(
        Paths,
        Sequences,
        Background,
        EmissionParameters,
        TransitionParameters,
        "nonhomo",
        fg_state,
        merge_neighbouring_sites,
        minimal_site_length,
        seq_file=EmissionParameters["dat_file_clip"],
        bck_file=EmissionParameters["dat_file_bg"],
    )

    Sequences = h5py.File(EmissionParameters["dat_file_clip"], "r")
    Background = h5py.File(EmissionParameters["dat_file_bg"], "r")

    print("Write peaks")
    # Write the results
    WriteResults(Sequences, Background, ScoredSites, OutFile, GeneAnnotation)

    generate_bed(OutFile, pv_cutoff=0.05)
    return


def generate_bed(file, pv_cutoff=0.05):
    df = pd.read_table(file, low_memory=False)

    # Determine the cutoff for the p-vales aften Bonferroni
    cutoff = np.log(0.05 / df.shape[0])
    df = df[df["pv"] < cutoff]

    # Make the names unique
    df["Gene"] = df["Gene"] + ["_" + str(i) for i in range(df.shape[0])]

    # Transform the Score to be betweeen 0 and 1000
    df["SiteScore"] /= np.max(df["SiteScore"]) * 0.001

    # Compute thick start and end
    df["ThickStart"] = 0
    df["ThickStop"] = 0

    df.loc[df["Strand"] == 1, "ThickStart"] = df["Start"] + df["max_pos"]
    df.loc[df["Strand"] == -1, "ThickStart"] = df["Stop"] - df["max_pos"]

    df.loc[df["Strand"] == 1, "ThickStop"] = df["Start"] + df["max_pos"] + 1
    df.loc[df["Strand"] == -1, "ThickStop"] = df["Stop"] - df["max_pos"] + 1

    # Rename the strands
    df.loc[df["Strand"] == 1, "Strand"] = "+"
    df.loc[df["Strand"] == -1, "Strand"] = "-"

    # Keep only the columns that are relevant for the bed-file
    df = df[
        [
            "ChrName",
            "Start",
            "Stop",
            "Gene",
            "SiteScore",
            "Strand",
            "ThickStart",
            "ThickStop",
        ]
    ].sort_values("SiteScore", ascending=False)

    # Write the output
    df.to_csv(file.replace(".txt", ".bed"), sep="\t", header=False, index=False)
    return


def GetSites(
    Paths,
    Sequences,
    Background,
    EmissionParameters,
    TransitionParameters,
    TransitionTypeFirst,
    fg_state,
    merge_neighbouring_sites,
    minimal_site_length,
    seq_file="",
    bck_file="",
):
    """Get the predicted sites."""
    # Extract the paths
    Sites = convert_paths_to_sites(
        Paths, fg_state, merge_neighbouring_sites, minimal_site_length
    )

    nr_of_genes = len(list(Sequences.keys()))
    gene_nr_dict = {}
    for i, curr_gene in enumerate(Sequences.keys()):
        gene_nr_dict[curr_gene] = i

    sites_keys = [key for key in list(Sites.keys()) if len(Sites[key]) > 0]
    f = lambda key, Sites=Sites, nr_of_genes=nr_of_genes, gene_nr_dict=gene_nr_dict, seq_file=seq_file, bck_file=bck_file, EmissionParameters=EmissionParameters, TransitionParameters=TransitionParameters, TransitionTypeFirst=TransitionTypeFirst, fg_state=fg_state, merge_neighbouring_sites=merge_neighbouring_sites, minimal_site_length=minimal_site_length: (
        Sites[key],
        key,
        nr_of_genes,
        gene_nr_dict[key],
        seq_file,
        bck_file,
        EmissionParameters,
        TransitionParameters,
        TransitionTypeFirst,
        fg_state,
        merge_neighbouring_sites,
        minimal_site_length,
    )

    data = map(f, sites_keys)

    LoadReads.close_data_handles(handles=[Sequences, Background])
    if EmissionParameters["nb_proc"] == 1:
        ScoredSites = dict([GetSitesForGene(curr_slice) for curr_slice in data])
    else:
        pool = multiprocessing.get_context("spawn").Pool(
            EmissionParameters["nb_proc"], maxtasksperchild=10
        )
        results = pool.imap(GetSitesForGene, data, chunksize=1)
        pool.close()
        pool.join()
        ScoredSites = dict([res for res in results])
    for key in list(ScoredSites.keys()):
        if len(ScoredSites[key]) == 0:
            del ScoredSites[key]

    return ScoredSites


def GetSitesForGene(data):
    """Determine the score of the sites for each gene."""
    # Computing the probabilities for the current gene
    (
        Sites,
        gene,
        nr_of_genes,
        gene_nr,
        seq_file,
        bck_file,
        EmissionParameters,
        TransitionParameters,
        TransitionTypeFirst,
        fg_state,
        merge_neighbouring_sites,
        minimal_site_length,
    ) = data

    # Turn the Sequence and Bacground objects into dictionaries again such that
    # the subsequent methods for using these do not need to be modified
    if len(Sites) == 0:
        return gene, []

    NrOfStates = EmissionParameters["NrOfStates"]

    Sites = dict([(gene, Sites)])

    Sequences = h5py.File(EmissionParameters["dat_file_clip"], "r")
    Background = h5py.File(EmissionParameters["dat_file_bg"], "r")

    Sequences_per_gene = PreloadSequencesForGene(Sequences, gene)
    Background_per_gene = PreloadSequencesForGene(Background, gene)

    Ix = GetModelIx(Sequences_per_gene, Type="all")

    if np.sum(Ix) == 0:
        return gene, []

    if EmissionParameters["filter_snps"]:
        Ix = GetModelIx(
            Sequences_per_gene,
            Type="no_snps_conv",
            snps_thresh=EmissionParameters["snps_thresh"],
            snps_min_cov=EmissionParameters["snps_min_cov"],
            Background=Background_per_gene,
        )
    else:
        Ix = GetModelIx(Sequences_per_gene, Type="Conv")

    # Only compute the emission probability for regions where a site is
    ix_sites = np.zeros_like(Ix)
    ix_sites_len = Ix.shape[0]
    for currsite in Sites[gene]:
        ix_sites[max(0, currsite[0] - 1) : min(ix_sites_len, currsite[1] + 1)] = 1
    ix_sites = ix_sites == 1

    # 2) Compute the probabilities for both states
    EmmisionProbGene = np.log(
        np.ones((NrOfStates, Ix.shape[0])) * (1 / np.float64(NrOfStates))
    )
    CurrStackSum = StackData(Sequences_per_gene)
    CurrStackVar = StackData(Sequences_per_gene, add="no")
    CurrStackSumBck = StackData(Background_per_gene, add="only_cov")

    CurrStackVarSumm = StackData(Sequences_per_gene, add="only_var_summed")
    EmmisionProbGeneDir = np.zeros_like(EmmisionProbGene)

    if EmissionParameters["glm_weight"] < 0.0:
        weight1 = 1.0
        weight2 = 1.0
    elif EmissionParameters["glm_weight"] == 0.0:
        weight1 = 0.0000001
        weight2 = 1.0 - weight1
    elif EmissionParameters["glm_weight"] == 1.0:
        weight1 = 0.9999999
        weight2 = 1.0 - weight1
    else:
        weight1 = EmissionParameters["glm_weight"]
        weight2 = 1.0 - EmissionParameters["glm_weight"]

    for State in range(NrOfStates):
        EmmisionProbGene[State, ix_sites] = np.log(
            weight1
        ) + emission_prob.predict_expression_log_likelihood_for_gene(
            CurrStackSum[:, ix_sites], State, nr_of_genes, gene_nr, EmissionParameters
        )
        if EmissionParameters["bg_type"] == "Coverage":
            EmmisionProbGene[State, ix_sites] += np.log(
                weight1
            ) + emission_prob.predict_expression_log_likelihood_for_gene(
                CurrStackSumBck[:, ix_sites],
                State,
                nr_of_genes,
                gene_nr,
                EmissionParameters,
                "bg",
            )
        if EmissionParameters["bg_type"] == "Coverage_bck":
            EmmisionProbGene[State, ix_sites] += np.log(
                weight1
            ) + emission_prob.predict_expression_log_likelihood_for_gene(
                CurrStackSumBck[:, ix_sites],
                State,
                nr_of_genes,
                gene_nr,
                EmissionParameters,
                "bg",
            )
        EmmisionProbGeneDir[State, Ix] = np.log(
            weight2
        ) + diag_event_model.pred_log_lik(
            CurrStackVar[:, Ix], State, EmissionParameters
        )
        EmmisionProbGene[State, Ix] += np.log(weight2) + EmmisionProbGeneDir[State, Ix]

    Counts = StackData(Sequences_per_gene, add="all")

    Score = EmmisionProbGene
    CurrStack = CurrStackVar
    # Compute the scores when staying in the same state
    # RowIx = list(range(16)) + list(range(17, 38)) + list(range(39,44))
    strand = Sequences_per_gene["strand"]

    # Get the coverages for the froeground and background
    CountsSeq = StackData(Sequences_per_gene, add="only_cov")
    CountsBck = StackData(Background_per_gene, add="only_cov")

    if strand == 0:
        strand = -1
    # Since we the transition probabilty is the same for all States we do not
    # need to compute it for the bayes factor this list contains the returned
    # sites
    sites = []
    for currsite in Sites[gene]:
        (
            mean_mat_fg,
            var_mat_fg,
            mean_mat_bg,
            var_mat_bg,
            counts_fg,
            counts_bg,
        ) = ComputeStatsForSite(
            CountsSeq,
            CountsBck,
            currsite,
            fg_state,
            nr_of_genes,
            gene_nr,
            EmissionParameters,
        )

        site = {}
        site["Start"] = currsite[0]
        site["Stop"] = currsite[1]
        site["Strand"] = strand
        site["SiteScore"] = EvaluateSite(Score, currsite, fg_state)
        site["Coverage"] = np.sum(
            np.sum(Counts[:, site["Start"] : site["Stop"]], axis=0)
        )
        site["Variants"] = np.sum(
            CurrStackVarSumm[:, site["Start"] : site["Stop"]], axis=1
        )
        site["mean_mat_fg"] = mean_mat_fg
        site["var_mat_fg"] = var_mat_fg
        site["mean_mat_bg"] = mean_mat_bg
        site["var_mat_bg"] = var_mat_bg
        site["counts_fg"] = counts_fg
        site["counts_bg"] = counts_bg

        p = mean_mat_fg / var_mat_fg
        n = (mean_mat_fg**2) / (var_mat_fg - mean_mat_fg)
        site["pv"] = nbinom.logsf(counts_fg, n, p)
        site["max_pos"] = get_max_position(Score, currsite, fg_state, strand)
        site["dir_score"] = EvaluateSite(EmmisionProbGeneDir, currsite, fg_state)
        if site["SiteScore"] < 0.0:
            continue
        sites.append(site)

    LoadReads.close_data_handles(handles=[Sequences, Background])

    return gene, sites


def convert_paths_to_sites(
    Paths, fg_state, merge_neighbouring_sites, minimal_site_length
):
    """Compute the site predictions using the paths."""
    sites = defaultdict(list)

    # Iterate over the paths
    for gene in Paths:
        curr_path = Paths[gene] == fg_state
        Starts = np.where(
            np.concatenate(
                ([curr_path[0]], curr_path[:-1] != curr_path[1:], [curr_path[-1]])
            )
        )[0][::2]
        Stops = np.where(
            np.concatenate(
                ([curr_path[0]], curr_path[:-1] != curr_path[1:], [curr_path[-1]])
            )
        )[0][1::2]
        nr_sites = Starts.shape[0]
        sites[gene] = [
            [Starts[i], Stops[i]]
            for i in range(nr_sites)
            if (Stops[i] - Starts[i] >= minimal_site_length)
        ]

    return sites


def ComputeStatsForSite(
    CountsSeq, CountsBck, Site, fg_state, nr_of_genes, gene_nr, EmissionParameters
):
    """Get the score for a Site."""
    Start = Site[0]
    Stop = Site[1]

    mean_mat_fg, var_mat_fg = emission_prob.get_expected_mean_and_var(
        CountsSeq[:, Start : (Stop + 1)],
        fg_state,
        nr_of_genes,
        gene_nr,
        EmissionParameters,
        curr_type="fg",
    )
    mean_mat_bg, var_mat_bg = emission_prob.get_expected_mean_and_var(
        CountsBck[:, Start : (Stop + 1)],
        fg_state,
        nr_of_genes,
        gene_nr,
        EmissionParameters,
        curr_type="bg",
    )

    mean_mat_fg = np.sum(np.sum(mean_mat_fg, axis=0))
    var_mat_fg = np.sum(np.sum(var_mat_fg, axis=0))
    mean_mat_bg = np.sum(np.sum(mean_mat_bg, axis=0))
    var_mat_bg = np.sum(np.sum(var_mat_bg, axis=0))
    counts_fg = np.sum(np.sum(CountsSeq[:, Start : (Stop + 1)], axis=0))
    counts_bg = np.sum(np.sum(CountsBck[:, Start : (Stop + 1)], axis=0))

    return mean_mat_fg, var_mat_fg, mean_mat_bg, var_mat_bg, counts_fg, counts_bg


def get_max_position(Score, Site, fg_state, strand):
    """Get the site where the score is maximal."""
    Start = Site[0]
    Stop = Site[1]
    ix_bg = list(range(Score.shape[0]))
    ix_bg.remove(fg_state)
    FGScore = Score[fg_state, Start : (Stop + 1)]
    AltScore = Score[ix_bg, Start : (Stop + 1)]

    norm = logsumexp(AltScore, axis=0)

    ix_ok = np.isinf(norm) + np.isnan(norm)
    if np.sum(ix_ok) < norm.shape[0]:
        SiteScore = FGScore[ix_ok == 0] - norm[ix_ok == 0]
    else:
        print("Score problematic")
        SiteScore = FGScore

    max_pos = np.int64(np.round(np.mean(np.where(SiteScore == np.max(SiteScore))[0])))
    if strand == -1:
        pos = Stop - Start - max_pos
    else:
        pos = max_pos

    return pos


def EvaluateSite(Score, Site, fg_state):
    """Get the score for a Site."""
    Start = Site[0]
    Stop = Site[1]
    ix_bg = list(range(Score.shape[0]))
    ix_bg.remove(fg_state)
    FGScore = np.sum(Score[fg_state, Start : (Stop + 1)])
    AltScore = np.sum(Score[ix_bg, Start : (Stop + 1)], axis=1)
    norm = logsumexp(AltScore)
    if not (np.isinf(norm) or np.isnan(norm)):
        SiteScore = FGScore - norm
    else:
        print("Score problematic")
        SiteScore = FGScore

    return SiteScore


def WriteResults(Sequences, Background, ScoredSites, OutFile, GeneAnnotation):
    """Write the sites into a result file."""
    # Get the gene annotation
    Iter = GeneAnnotation.features_of_type("gene")
    Genes = []
    for gene in Iter:
        Genes.append(gene)

    # Print the results
    fid = open(OutFile, "w")
    VariantList = [
        Nucl_1 + "-" + Nucl_2
        for Nucl_1 in ["A", "C", "G", "T"]
        for Nucl_2 in ["A", "C", "G", "T", "D"]
    ]
    Header = (
        "\t".join(
            ["Gene", "ChrName", "Start", "Stop", "Strand", "SiteScore", "Coverage"]
            + VariantList
            + [
                "mean_mat_fg",
                "var_mat_fg",
                "mean_mat_bg",
                "var_mat_bg",
                "counts_fg",
                "counts_bg",
                "pv",
                "max_pos",
                "dir_score",
            ]
        )
        + "\n"
    )
    fid.write(Header)
    for gene in Genes:
        gene_name = gene.id
        if gene_name not in ScoredSites:
            continue
        # Transform the Coordinates
        for site in ScoredSites[gene_name]:
            # Process the current site
            CurrLine = (
                "\t".join(
                    [
                        gene_name,
                        gene.chrom,
                        str(GetGenomicCoord(gene, site["Start"])),
                        str(GetGenomicCoord(gene, site["Stop"])),
                        str(site["Strand"]),
                        str(site["SiteScore"]),
                        str(site["Coverage"]),
                    ]
                    + [
                        str(int(site["Variants"][i]))
                        for i in range(len(site["Variants"]))
                    ]
                    + [
                        str(site["mean_mat_fg"]),
                        str(site["var_mat_fg"]),
                        str(site["mean_mat_bg"]),
                        str(site["var_mat_bg"]),
                        str(site["counts_fg"]),
                        str(site["counts_bg"]),
                        str(site["pv"]),
                        str(site["max_pos"]),
                        str(site["dir_score"]),
                    ]
                )
                + "\n"
            )

            fid.write(CurrLine)
    fid.close()

    return


def GetGenomicCoord(gene, Coord):
    """Return formatted gene coordinates."""
    return gene.start + Coord


def estimate_library_size(Sequences):
    """Estimate the library size of all samples."""
    lib_size_dict = defaultdict(list)
    lib_size_red = defaultdict(int)
    # Get the gene expressions
    for gene in list(Sequences.keys()):
        for key in list(Sequences[gene]["Coverage"].keys()):
            lib_size_dict[key].append(Sequences[gene]["Coverage"][key][()].sum())

    # Compute the (weighted) median of non zero genes
    for key in lib_size_dict:
        exprs = np.array([g for g in lib_size_dict[key] if g > 0])

        lib_size_red[key] = np.median(exprs)

    return lib_size_red


def ParallelGetMostLikelyPath(
    MostLikelyPaths,
    Sequences,
    Background,
    EmissionParameters,
    TransitionParameters,
    TransitionTypeFirst,
    RandomNoise=False,
    chunksize=1,
    verbosity=1,
):
    """Compute the most likely path."""

    for gene in list(MostLikelyPaths.keys()):
        del MostLikelyPaths[gene]
    nr_of_genes = len(list(Sequences.keys()))
    gene_nr_dict = {}
    for i, curr_gene in enumerate(Sequences.keys()):
        gene_nr_dict[curr_gene] = i

    print("Computing most likely path")
    t = time.time()

    data = zip(
        list(Sequences.keys()),
        itertools.repeat(nr_of_genes),
        list(gene_nr_dict.values()),
        itertools.repeat(EmissionParameters),
        itertools.repeat(TransitionParameters),
        itertools.repeat(TransitionTypeFirst),
        itertools.repeat(RandomNoise),
    )

    LoadReads.close_data_handles(handles=[Sequences, Background])

    if EmissionParameters["nb_proc"] == 1:
        results = [ParallelGetMostLikelyPathForGene(curr_slice) for curr_slice in data]
    else:
        print("Spawning processes")
        pool = multiprocessing.get_context("spawn").Pool(
            EmissionParameters["nb_proc"], maxtasksperchild=5
        )
        results = pool.imap(ParallelGetMostLikelyPathForGene, data, chunksize)
        pool.close()
        pool.join()
        print("Collecting results")
        results = [res for res in results]

    MostLikelyPaths = dict(
        zip([result[0] for result in results], [result[1] for result in results])
    )

    # Compute the logliklihood of the gene
    LogLikelihood = sum([result[2] for result in results])
    del results

    if verbosity > 0:
        print("\nDone: Elapsed time: " + str(time.time() - t))

    return MostLikelyPaths, LogLikelihood


def ParallelGetMostLikelyPathForGene(data):
    """Compute the most likely path for a gene."""
    (
        gene,
        nr_of_genes,
        gene_nr,
        EmissionParameters,
        TransitionParameters,
        TransitionTypeFirst,
        RandomNoise,
    ) = data

    # Turn the Sequence and Bacground objects into dictionaries again such that
    # the subsequent methods for using these do not need to be modified
    Sequences = h5py.File(EmissionParameters["dat_file_clip"], "r")
    Background = h5py.File(EmissionParameters["dat_file_bg"], "r")

    # Parse the parameters
    alpha = EmissionParameters["Diag_event_params"]
    PriorMatrix = EmissionParameters["PriorMatrix"]
    NrOfStates = EmissionParameters["NrOfStates"]

    fg_state, bg_state = emission_prob.get_fg_and_bck_state(
        EmissionParameters, final_pred=True
    )
    # Score the state sequences
    # 1) Determine the positions where an observation is possible

    Sequences_per_gene = PreloadSequencesForGene(Sequences, gene)
    Background_per_gene = PreloadSequencesForGene(Background, gene)

    Ix = GetModelIx(Sequences_per_gene, Type="all")

    if np.sum(Ix) == 0:
        CurrPath = 2 * np.ones((0, Ix.shape[0]), dtype=np.int64)
        return [gene, CurrPath, 0]

    if EmissionParameters["filter_snps"]:
        Ix = GetModelIx(
            Sequences_per_gene,
            Type="no_snps_conv",
            snps_thresh=EmissionParameters["snps_thresh"],
            snps_min_cov=EmissionParameters["snps_min_cov"],
            Background=Background_per_gene,
        )
    else:
        Ix = GetModelIx(Sequences_per_gene)

    # 2) Compute the probabilities for both states
    EmmisionProbGene = np.ones((NrOfStates, Ix.shape[0])) * (1 / np.float64(NrOfStates))

    CurrStackSum = StackData(Sequences_per_gene)
    CurrStackVar = StackData(Sequences_per_gene, add="no")
    CurrStackSumBck = StackData(Background_per_gene, add="only_cov")

    if EmissionParameters["glm_weight"] < 0.0:
        weight1 = 1.0
        weight2 = 1.0
    elif EmissionParameters["glm_weight"] == 0.0:
        weight1 = 0.0000001
        weight2 = 1.0 - weight1
    elif EmissionParameters["glm_weight"] == 1.0:
        weight1 = 0.9999999
        weight2 = 1.0 - weight1
    else:
        weight1 = EmissionParameters["glm_weight"]
        weight2 = 1.0 - EmissionParameters["glm_weight"]

    for State in range(NrOfStates):
        if not EmissionParameters["ign_GLM"]:
            if isinstance(EmissionParameters["ExpressionParameters"][0], np.ndarray):
                EmmisionProbGene[State, :] = np.log(
                    weight1
                ) + emission_prob.predict_expression_log_likelihood_for_gene(
                    CurrStackSum, State, nr_of_genes, gene_nr, EmissionParameters
                )
                if EmissionParameters["bg_type"] == "Coverage":
                    EmmisionProbGene[State, :] += np.log(
                        weight1
                    ) + emission_prob.predict_expression_log_likelihood_for_gene(
                        CurrStackSumBck,
                        State,
                        nr_of_genes,
                        gene_nr,
                        EmissionParameters,
                        "bg",
                    )
                if EmissionParameters["bg_type"] == "Coverage_bck":
                    EmmisionProbGene[State, :] += np.log(
                        weight1
                    ) + emission_prob.predict_expression_log_likelihood_for_gene(
                        CurrStackSumBck,
                        State,
                        nr_of_genes,
                        gene_nr,
                        EmissionParameters,
                        "bg",
                    )
        if not EmissionParameters["ign_diag"]:
            EmmisionProbGene[State, Ix] += np.log(
                weight2
            ) + diag_event_model.pred_log_lik(
                CurrStackVar[:, Ix], State, EmissionParameters
            )
        if State == fg_state:
            if EmissionParameters["LastIter"]:
                EmmisionProbGene[State, :] -= EmissionParameters["fg_pen"]
    if RandomNoise:
        EmmisionProbGene = np.logaddexp(
            EmmisionProbGene,
            np.random.uniform(
                np.min(EmmisionProbGene[np.isfinite(EmmisionProbGene)]) - 4,
                np.min(EmmisionProbGene[np.isfinite(EmmisionProbGene)]) - 0.1,
                EmmisionProbGene.shape,
            ),
        )  # Add some random noise

    # Get the transition probabilities
    TransistionProbabilities = np.float64(
        np.tile(np.log(TransitionParameters[0]), (EmmisionProbGene.shape[1], 1, 1)).T
    )

    CurrPath, Currloglik = viterbi.viterbi(
        np.float64(EmmisionProbGene),
        TransistionProbabilities,
        np.float64(np.log(PriorMatrix)),
    )
    CurrPath = np.int8(CurrPath)

    del (
        TransistionProbabilities,
        EmmisionProbGene,
        CurrStackSum,
        CurrStackVar,
        CurrStackSumBck,
        Ix,
    )

    LoadReads.close_data_handles(handles=[Sequences, Background])

    return [gene, CurrPath, Currloglik]


def subsample_suff_stat(Counts, NrOfCounts, subsample_size=250000):
    """Create a subsample of the counts."""
    # Iterate over the keys
    for key in Counts:
        # Determine the new sample
        if NrOfCounts[key].shape[0] == 0:
            # Try to rescue parameter fitting by adding some data from other
            # regions for fitting.
            for temp_key in Counts:
                if NrOfCounts[temp_key].shape[0] > 10:
                    NrOfCounts[key] = np.ones((1, 10))
                    ix = np.random.randint(NrOfCounts[temp_key].shape[0], size=10)
                    Counts[key] = Counts[temp_key][:, ix]
                    break

        new_counts = np.random.multinomial(
            min(subsample_size, np.sum(NrOfCounts[key][0, :])),
            NrOfCounts[key][0, :] / np.float64(np.sum(NrOfCounts[key][0, :])),
            size=1,
        )

        # Check that not more sample than orignialy existing were present
        ix = new_counts > NrOfCounts[key]
        new_counts[ix] = NrOfCounts[key][ix]

        # If no count is there take the original sample
        if np.sum(new_counts) == 0:
            new_counts = NrOfCounts[key][0, :]

        ix_non_zero = new_counts > 0
        temp_counts = np.zeros((1, np.sum(new_counts > 0)))
        temp_counts[0, :] = new_counts[ix_non_zero]
        NrOfCounts[key] = temp_counts
        Counts[key] = Counts[key][:, ix_non_zero[0, :]]

    return Counts, NrOfCounts



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['omniCLIP', 'omniCLIP.data_parsing', 'omniCLIP.omni_stat']

package_data = \
{'': ['*']}

install_requires = \
['Cython>=0.29.28,<0.30.0',
 'biopython>=1.79,<2.0',
 'gffutils>=0.10.1,<0.11.0',
 'h5py>=3.6.0,<4.0.0',
 'intervaltree>=3.1.0,<4.0.0',
 'matplotlib>=3.5.1,<4.0.0',
 'numpy>=1.22.3,<2.0.0',
 'pandas>=1.4.2,<2.0.0',
 'pysam>=0.19.0,<0.20.0',
 'scipy>=1.8.0,<2.0.0',
 'sklearn>=0.0,<0.1',
 'statsmodels>=0.13.2,<0.14.0']

entry_points = \
{'console_scripts': ['omniCLIP = omniCLIP.omniCLIP:main']}

setup_kwargs = {
    'name': 'omniclip',
    'version': '0.2.6',
    'description': '',
    'long_description': '# omniCLIP\nomniCLIP is a Bayesian peak caller that can be applied to data from CLIP-Seq experiments to detect regulatory elements in RNAs.\n\n## Overview\n\n[Introduction](#introduction)\n\n[Dependencies](#dependencies)\n\n[Installation](#installation)\n\n[Usage](#usage)\n\n[Contributors](#contributors)\n\n[License](#license)\n\n\n## Introduction\nomniCLIP can call peaks for CLIP-Seq data data while accounting for confounding factors such as the gene expression and it automatically learns relevant diagnostic events from the data. Furtermore, it can leverage replicate information and model technical and biological variance.\n\n## Dependencies and Requirements\nomniCLIP requires Python (v.3.7) and the libraries described in the `environment.yml` file. All required dependencies can be installed using *conda* by executing the following in the main project directory :\n```\n$ conda env create -f environment.yml\n```\n\nThe environment then needs to be activated in order to run omniCLIP :\n```\n$ conda activate omniEnv\n```\n\n## Installation\n\n### Manual installation\nThe latest stable release in the ***master*** branch can be downloaded by executing:\n```\n$ git clone -b master https://github.com/simojoe/omniCLIP.git\n```\nAfter this the following command has to be executed:\n```\n$ python3 setup.py\n```\nThis compiles the Cython code for the viterbi algorithm.\n\n## Usage\nAn omniCLIP analysis is run into four different steps :\n- Generating the annotation database\n- Parsing the background RNA-seq files\n- Parsing the CLIP files\n- Running the omniCLIP algorithm\n\nThe following is an example of the commands. The commands in this example only show the **required** arguments for the analysis. The following files are necessary to run an analysis.\n\nFile name  | Description\n------------- | -------------\n$GFF_file | Genome annotation file\n$GENOME_dir | Directory containing FASTA files with each of the chromosomes sequence\n$BG_file[1,2,...] | BAM files for the background library. The alignments need to have the MD and NM tags.\n$CLIP_file[1,2,...] | BAM files for the CLIP library. The alignments need to have the MD and NM tags.\n\nThe following files will be created.\n\nFile name  | Description\n------------- | -------------\n$DB_file | SQL database of the genome annotation.\n$BG_dat | H5PY file of the parsed background library.\n$CLIP_dat | H5PY file of the parsed CLIP library.\n$OUT_dir | Directory containing the final results\n\n\n### 1. Generating the annotation database\n```\n$ python3 omniCLIP.py generateDB \\\n--gff-file $GFF_file --db-file $DB_file\n```\n\n### 2. Parsing the background RNA-seq files\n```\n$ python3 omniCLIP.py parsingBG \\\n--db-file $DB_file --genome-dir $GENOME_dir \\\n--bg-files $BG_file1 --bg-files $BG_file2 \\\n--out-file $BG_dat\n```\n\n### 3. Parsing the CLIP files\n```\n$ python3 omniCLIP.py parsingCLIP \\\n--db-file $DB_file --genome-dir $GENOME_dir \\\n--clip-files $CLIP_file1 --clip-files $CLIP_file2 \\\n--out-file $CLIP_dat\n```\n\n\n### 4. Running the omniCLIP algorithm\n```\n$ python3 omniCLIP.py run_omniCLIP \\\n--db-file $DB_file --bg-dat $BG_dat --clip-dat $CLIP_dat \\\n--out-dir $OUT_dir\n```\n\n### Optional arguments\nFor the full list of optional arguments of the different step, consult the help of the commands using :\n```\n$ python3 omniCLIP.py parsingBG --help\n```\n\n## Contributors\n\n\n\n## License\nGNU GPL license (v3)\n',
    'author': 'Philipp Boss',
    'author_email': 'philipp.drewe@googlemail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/y9c/omniCLIP',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



"""omniCLIP is a CLIP-Seq peak caller.

Copyright (C) 2017 Philipp Boss

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

__version__ = "0.2.6"

import argparse
import gc
import os
import time

import gffutils
import h5py
import numpy as np

from .data_parsing import (
    CreateGeneAnnotDB,
    LoadReads,
    ParsingArgs,
    ParsingPositions,
    tools,
)
from .omni_stat import emission_prob, mixture_tools, trans
from .omni_stat.utils import get_mem_usage


def run_omniCLIP(args):
    """Run omniCLIP function."""
    # Parsing the arguments
    EmissionParameters = ParsingArgs.parsing_argparse(args)

    # Creating temp copies of Sequence and Background
    ParsingArgs.dup_seqfiles(EmissionParameters)

    # Parsing arguments dependents of Sequence and Background
    EmissionParameters = ParsingArgs.parsing_files(args, EmissionParameters)

    # Load the gene annotation
    print("Loading gene annotation")
    # if args.gene_anno_file.split(".")[-1] == "db":
    GeneAnnotation = gffutils.FeatureDB(args.gene_anno_file, keep_order=True)

    import warnings

    warnings.filterwarnings("error")

    # Load the reads
    get_mem_usage(EmissionParameters["verbosity"])
    t = time.time()
    Sequences = h5py.File(EmissionParameters["dat_file_clip"], "r+")
    Background = h5py.File(EmissionParameters["dat_file_bg"], "r+")

    msg = "Done: Elapsed time: " + str(time.time() - t)
    get_mem_usage(EmissionParameters["verbosity"], t=t, msg=msg)

    # Initializing parameters
    print("Initialising the parameters")
    TransitionParameters = [EmissionParameters["TransMat"], []]

    # Transistion parameters
    IterParameters = [EmissionParameters, TransitionParameters]

    # Start computation

    # Iteratively fit the parameters of the model
    OldLogLikelihood = 0
    CurrLogLikelihood = -np.inf
    CurrIter = 0
    LoglikelihodList = []
    First = 1
    Paths = {}
    iter_cond = True

    while iter_cond:
        print("\n")
        print("Iteration: " + str(CurrIter))
        if EmissionParameters["verbosity"] > 1:
            print(IterParameters[0])

        OldLogLikelihood = CurrLogLikelihood

        CurrLogLikelihood, IterParameters, First, Paths = PerformIteration(
            Sequences,
            Background,
            IterParameters,
            EmissionParameters["NrOfStates"],
            First,
            Paths,
            verbosity=EmissionParameters["verbosity"],
        )

        gc.collect()

        if EmissionParameters["verbosity"] > 1:
            print("Log-likelihood: " + str(CurrLogLikelihood))
        LoglikelihodList.append(CurrLogLikelihood)

        if EmissionParameters["verbosity"] > 1:
            print(LoglikelihodList)
        CurrIter += 1

        if CurrIter >= EmissionParameters["max_it"]:
            print("Maximal number of iterations reached")

        if CurrIter < max(3, EmissionParameters["max_it"]):
            iter_cond = True
        else:
            iter_cond = (
                (CurrIter < EmissionParameters["max_it"])
                and (
                    (
                        abs(CurrLogLikelihood - OldLogLikelihood)
                        / max(abs(CurrLogLikelihood), abs(OldLogLikelihood))
                    )
                    > 0.01
                )
                and (
                    abs(CurrLogLikelihood - OldLogLikelihood) > args.tol_lg_lik
                )
            )

    # Return the fitted parameters
    print("Finished parameter fitting")

    EmissionParameters, TransitionParameters = IterParameters
    if not isinstance(
        EmissionParameters["ExpressionParameters"][0], np.ndarray
    ):
        print("Emmision parameters have not been fit yet")
        return

    OutFile = os.path.join(
        EmissionParameters["out_dir"],
        EmissionParameters["out_file_base"] + ".txt",
    )

    # Determine which state has higher weight in fg.
    get_mem_usage(EmissionParameters["verbosity"])

    fg_state, bg_state = emission_prob.get_fg_and_bck_state(
        EmissionParameters, final_pred=True
    )
    if EmissionParameters["fg_pen"] > 0.0:
        print("Recomputing paths")
        EmissionParameters["LastIter"] = True
        Sequences = h5py.File(EmissionParameters["dat_file_clip"], "r")
        Background = h5py.File(EmissionParameters["dat_file_bg"], "r")

        Paths, LogLike = tools.ParallelGetMostLikelyPath(
            Paths,
            Sequences,
            Background,
            EmissionParameters,
            TransitionParameters,
            "nonhomo",
            verbosity=EmissionParameters["verbosity"],
        )

        Sequences = h5py.File(EmissionParameters["dat_file_clip"], "r")
        Background = h5py.File(EmissionParameters["dat_file_bg"], "r")

    tools.GeneratePred(
        Paths,
        Sequences,
        Background,
        IterParameters,
        GeneAnnotation,
        OutFile,
        fg_state,
        bg_state,
        pv_cutoff=EmissionParameters["pv_cutoff"],
        verbosity=EmissionParameters["verbosity"],
    )
    print("Done")

    # Remove the temporary files
    print("Removing temporary files")
    os.remove(EmissionParameters["dat_file_clip"])
    os.remove(EmissionParameters["dat_file_bg"])

    return


def PerformIteration(
    Sequences,
    Background,
    IterParameters,
    NrOfStates,
    First,
    NewPaths={},
    verbosity=1,
):
    """This function performs an iteration of the HMM algorithm."""
    # Unpack the Iteration parameters
    EmissionParameters = IterParameters[0]
    TransitionParameters = IterParameters[1]

    # Get new most likely path
    if First:
        NewPaths, LogLike = tools.ParallelGetMostLikelyPath(
            NewPaths,
            Sequences,
            Background,
            EmissionParameters,
            TransitionParameters,
            "homo",
            RandomNoise=True,
            verbosity=verbosity,
        )

        Sequences = h5py.File(EmissionParameters["dat_file_clip"], "r")
        Background = h5py.File(EmissionParameters["dat_file_bg"], "r")

        get_mem_usage(verbosity)

    # Perform EM to compute the new emission parameters
    print("Fitting emission parameters")
    get_mem_usage(verbosity)

    NewEmissionParameters = FitEmissionParameters(
        Sequences,
        Background,
        NewPaths,
        EmissionParameters,
        First,
        verbosity=verbosity,
    )

    if First:
        First = 0

    get_mem_usage(verbosity)

    # Fit the transition matrix parameters
    NewTransitionParameters = TransitionParameters
    print("Fitting transition parameters")
    get_mem_usage(verbosity)

    LoadReads.close_data_handles(handles=[Sequences, Background])
    Sequences = h5py.File(EmissionParameters["dat_file_clip"], "r")
    Background = h5py.File(EmissionParameters["dat_file_bg"], "r")

    TransistionPredictors = trans.FitTransistionParameters(
        Sequences,
        Background,
        TransitionParameters,
        NewPaths,
        verbosity=verbosity,
    )

    NewTransitionParameters[1] = TransistionPredictors
    get_mem_usage(verbosity)

    NewIterParameters = [NewEmissionParameters, NewTransitionParameters]

    print("Computing most likely path")
    get_mem_usage(verbosity)

    gc.collect()
    NewPaths, LogLike = tools.ParallelGetMostLikelyPath(
        NewPaths,
        Sequences,
        Background,
        EmissionParameters,
        TransitionParameters,
        "nonhomo",
        verbosity=verbosity,
    )

    Sequences = h5py.File(EmissionParameters["dat_file_clip"], "r")
    Background = h5py.File(EmissionParameters["dat_file_bg"], "r")

    CurrLogLikelihood = LogLike
    get_mem_usage(verbosity)
    if verbosity > 1:
        print("LogLik:")
        print(CurrLogLikelihood)
    return CurrLogLikelihood, NewIterParameters, First, NewPaths


def FitEmissionParameters(
    Sequences, Background, NewPaths, OldEmissionParameters, First, verbosity=1
):
    """Fit EmissionParameters."""
    print("Fitting emission parameters")
    t = time.time()
    # Unpack the arguments
    OldAlpha = OldEmissionParameters["Diag_event_params"]
    NrOfStates = OldEmissionParameters["NrOfStates"]
    OldPriorMatrix = OldEmissionParameters["PriorMatrix"]
    NewEmissionParameters = OldEmissionParameters

    # Compute new prior matrix
    PriorMatrix = np.zeros_like(OldPriorMatrix)
    for State in range(NrOfStates):
        for path in NewPaths:
            PriorMatrix[State] += np.sum(NewPaths[path] == State)

    # Check if one of the states is not used and add pseudo gene to prevent
    # singularities during distribution fitting
    if np.sum(PriorMatrix == 0) > 0:
        LoadReads.close_data_handles(handles=[Sequences, Background])
        Sequences = h5py.File(NewEmissionParameters["dat_file_clip"], "r+")
        Background = h5py.File(NewEmissionParameters["dat_file_bg"], "r+")
        Sequences, Background, NewPaths = add_pseudo_gene(
            Sequences, Background, NewPaths, PriorMatrix
        )
        print("Adds pseudo gene to prevent singular matrix during GLM fitting")

    CorrPriorMatrix = np.copy(PriorMatrix)
    CorrPriorMatrix[CorrPriorMatrix == 0] = (
        np.min(CorrPriorMatrix[CorrPriorMatrix > 0]) / 10
    )
    CorrPriorMatrix /= np.sum(CorrPriorMatrix)
    # Keep a copy to check which states are not used
    NewEmissionParameters["PriorMatrix"] = CorrPriorMatrix

    # Add Pseudo gene to Sequences, Background and Paths
    if NewEmissionParameters["ExpressionParameters"][0] is not None:
        Sequences, Background, NewPaths = add_pseudo_gene(
            Sequences, Background, NewPaths, PriorMatrix
        )

    # Compute parameters for the expression
    Sequences = h5py.File(NewEmissionParameters["dat_file_clip"], "r")
    if (NewEmissionParameters["bg_type"] != "None") and not First:
        if "Pseudo" in list(Sequences.keys()):
            nr_of_genes = len(list(Sequences.keys()))
            new_pars = NewEmissionParameters["ExpressionParameters"][0]
            new_pars = np.vstack(
                (
                    new_pars[:(nr_of_genes), :],
                    np.mean(new_pars[:(nr_of_genes), :]),
                    new_pars[(nr_of_genes):, :],
                )
            )
            NewEmissionParameters["ExpressionParameters"][0] = new_pars
    print("Estimating expression parameters")
    get_mem_usage(verbosity)

    NewEmissionParameters = emission_prob.estimate_expression_param(
        (NewEmissionParameters, NewPaths), verbosity=verbosity
    )

    get_mem_usage(verbosity)

    Sequences = h5py.File(NewEmissionParameters["dat_file_clip"], "r")
    Background = h5py.File(NewEmissionParameters["dat_file_bg"], "r")

    if NewEmissionParameters["bg_type"] != "None":
        if "Pseudo" in list(Sequences.keys()):
            nr_of_genes = len(list(Sequences.keys()))
            new_pars = NewEmissionParameters["ExpressionParameters"][0]
            new_pars = np.vstack(
                (new_pars[: (nr_of_genes - 1), :], new_pars[(nr_of_genes):, :])
            )
            NewEmissionParameters["ExpressionParameters"][0] = new_pars

    if NewEmissionParameters["skip_diag_event_mdl"] is False:
        # Compute parameters for the ratios
        print("Computing sufficient statistic for fitting md")
        get_mem_usage(verbosity)

        SuffStat = tools.GetSuffStat(
            NewPaths,
            NrOfStates,
            Type="Conv",
            EmissionParameters=NewEmissionParameters,
            verbosity=verbosity,
        )

        # Vectorize SuffStat
        Counts, NrOfCounts = tools.ConvertSuffStatToArrays(SuffStat)

        del SuffStat
        get_mem_usage(verbosity)
        if NewEmissionParameters["subs"]:
            Counts, NrOfCounts = tools.subsample_suff_stat(Counts, NrOfCounts)

        print("Fitting md distribution")
        get_mem_usage(verbosity)
        if NewEmissionParameters["diag_bg"]:
            print("Adjusting background")
            SuffStatBck = tools.GetSuffStatBck(
                NewPaths,
                NrOfStates,
                Type="Conv",
                EmissionParameters=NewEmissionParameters,
                verbosity=verbosity,
            )

            # Vectorize SuffStat
            CountsBck, NrOfCountsBck = tools.ConvertSuffStatToArrays(
                SuffStatBck
            )

            if NewEmissionParameters["subs"]:
                CountsBck, NrOfCountsBck = tools.subsample_suff_stat(
                    CountsBck, NrOfCountsBck
                )

            # Overwrite counts in other bins
            fg_state, bg_state = emission_prob.get_fg_and_bck_state(
                NewEmissionParameters, final_pred=True
            )

            for curr_state in list(Counts.keys()):
                if curr_state != fg_state:
                    Counts[curr_state] = CountsBck[fg_state]
                    NrOfCounts[curr_state] = NrOfCountsBck[fg_state]

            del SuffStatBck

        NewEmissionParameters = mixture_tools.em(
            Counts,
            NrOfCounts,
            NewEmissionParameters,
            x_0=OldAlpha,
            First=First,
            verbosity=verbosity,
        )

        get_mem_usage(verbosity)
        del Counts, NrOfCounts

    if "Pseudo" in list(Sequences.keys()):
        del Sequences["Pseudo"]
        del Background["Pseudo"]
        del NewPaths["Pseudo"]

    if verbosity > 0:
        print("Done: Elapsed time: " + str(time.time() - t))
    return NewEmissionParameters


def add_pseudo_gene(Sequences, Background, NewPaths, PriorMatrix):
    """Add pseudo gene to Sequences and Background."""
    nr_of_genes_to_gen = np.sum(PriorMatrix == 0)

    # If no pseudo gen has tp be generated, continue
    if nr_of_genes_to_gen == 0:
        return Sequences, Background, NewPaths

    # If pseudo is already in Sequences, skip
    if "Pseudo" in list(Sequences.keys()):
        return Sequences, Background, NewPaths

    # Generate pseudo genes
    # Get the gene lengths
    gen_lens = [
        Sequences[gene]["Coverage"]["0"][()].shape[1] for gene in Sequences
    ]

    random_gen = np.random.choice(
        np.arange(len(gen_lens)),
        1,
        p=np.array(gen_lens) / np.float64(sum(gen_lens)),
    )

    if len(random_gen) == 1:
        gene_name = list(Sequences.keys())[random_gen[0]]
    else:
        gene_name = list(Sequences.keys())[random_gen]

    Sequences["Pseudo"] = Sequences[gene_name]
    Background["Pseudo"] = Background[gene_name]

    zero_states = [i for i in range(len(PriorMatrix))]

    new_path = np.random.choice(
        zero_states, size=NewPaths[gene_name].shape, replace=True
    )
    NewPaths["Pseudo"] = new_path

    return Sequences, Background, NewPaths


def generateDB(args):
    """Given a GFF file, launches to CreateDB function.

    By default, omniCLIP only considers entries with a 'gene' feature. To
    include other features, as needed when using an Ensembl GFF3 for example,
    one needs to specify them at the command line.

    Example : --gene-features=gene,ncRNA
    """
    if args.gene_features is None:
        gene_features = ["gene"]
    else:
        gene_features = args.gene_features[0].split(",")

    # Creating the DB
    CreateGeneAnnotDB.CreateDB(args.gff_file, args.db_file, gene_features)


def parsingBG(args):
    """Parse the background (BG) BAM files."""
    GeneAnnotation = gffutils.FeatureDB(args.db_file, keep_order=True)
    LoadReads.load_data(
        bam_files=args.bg_libs,
        genome_dir=args.genome_dir,
        gene_annotation=GeneAnnotation,
        out_file=args.out_file,
        Collapse=args.collapsed,
        OnlyCoverage=args.only_coverage,
        mask_flank_variants=args.mask_flank_variants,
        max_mm=args.max_mm,
        ign_out_rds=args.ign_out_rds,
        rev_strand=args.rev_strand,
    )


def parsingCLIP(args):
    """Parse the CLIP (BG) BAM files."""
    GeneAnnotation = gffutils.FeatureDB(args.db_file, keep_order=True)
    LoadReads.load_data(
        bam_files=args.clip_libs,
        genome_dir=args.genome_dir,
        gene_annotation=GeneAnnotation,
        out_file=args.out_file,
        Collapse=args.collapsed,
        mask_flank_variants=args.mask_flank_variants,
        max_mm=args.max_mm,
        ign_out_rds=args.ign_out_rds,
        rev_strand=args.rev_strand,
        CLIP_exp=True,
        min_coverage=int(args.min_coverage),
        min_peak=int(args.min_peak),
    )

    Sequences = LoadReads.get_data_handle(args.out_file, write=True)

    if args.mask_miRNA:
        print("Removing miRNA-coverage")
        ParsingPositions.mask_miRNA_positions(Sequences, GeneAnnotation)

    if args.mask_ovrlp:
        print("Masking overlapping positions")
        ParsingPositions.mask_overlapping_positions(Sequences, GeneAnnotation)


def main():
    # with open("version.py") as fp:
    #    exec(fp.read(), version)

    parser = argparse.ArgumentParser(
        prog="PROG",
        description="omniCLIP - probabilistic identification of protein-RNA interactions from CLIP-seq data.\n\n"
        + "This is fork version by Chang Y.\n"
        + "Please cite the original paper: Drewe-Boss, P., Wessels, HH. & Ohler, U. omniCLIP: probabilistic identification of protein-RNA interactions from CLIP-seq data. Genome Biol 19, 183 (2018). https://doi.org/10.1186/s13059-018-1521-2",
    )
    parser.add_argument("--version", action="version", version=__version__)
    subparsers = parser.add_subparsers(
        title="subcommands", help="sub-command help", dest="command"
    )

    # Create the parser for the generateDB command
    parser_generateDB = subparsers.add_parser(
        "generateDB",
        help="generateDB help",
        description="Preprocessing of a GFF annotation file into an SQL database.",
    )
    parser_generateDB_reqNamed = parser_generateDB.add_argument_group(
        "required arguments"
    )
    parser_generateDB_reqNamed.add_argument(
        "--gff-file",
        dest="gff_file",
        help="Path to the .GFF annotation file",
        required=True,
    )
    parser_generateDB_reqNamed.add_argument(
        "--db-file",
        dest="db_file",
        help="Path to the output .GFF.DB file",
        required=True,
    )
    # Optional args for the generateDB command
    parser_generateDB.add_argument(
        "--gene_features",
        nargs="+",
        dest="gene_features",
        help="Features in the annotation file to be considered as genes",
        required=False,
    )

    # Shared optional arguments from parsingBG and parsingCLIP command
    parent_parsing = argparse.ArgumentParser(add_help=False)
    parent_parsing.add_argument(
        "--rev_strand",
        action="store",
        dest="rev_strand",
        choices=[0, 1],
        help="Only consider reads on the forward (0) or reverse strand (1) relative to the gene orientation",
        type=int,
        default=None,
    )
    parent_parsing.add_argument(
        "--collapsed",
        action="store_true",
        default=False,
        dest="collapsed",
        help="Reads are collapsed",
    )
    parent_parsing.add_argument(
        "--ign_out_rds",
        action="store_true",
        dest="ign_out_rds",
        help="ignore reads where the ends map ouside of the genes",
        default=False,
    )
    parent_parsing.add_argument(
        "--max-mismatch",
        action="store",
        dest="max_mm",
        help="Maximal number of mismatches that is allowed per read (default: 2)",
        type=int,
        default=2,
    )
    parent_parsing.add_argument(
        "--mask_flank_mm",
        action="store",
        dest="mask_flank_variants",
        help="Do not consider mismatches in the N bp at the ends of reads for diagnostic event modelling (default: 3)",
        type=int,
        default=3,
    )

    # Create the parser for the parsingBG command
    parser_parsingBG = subparsers.add_parser(
        "parsingBG",
        help="parsingBG help",
        description="Parsing the background files.",
        parents=[parent_parsing],
    )
    parser_parsingBG_reqNamed = parser_parsingBG.add_argument_group(
        "required arguments"
    )
    parser_parsingBG_reqNamed.add_argument(
        "--bg-files",
        action="append",
        dest="bg_libs",
        help="BAM files for background libraries",
        required=True,
    )
    parser_parsingBG_reqNamed.add_argument(
        "--db-file",
        action="store",
        dest="db_file",
        help="Path to the .GFF.DB file",
        required=True,
    )
    parser_parsingBG_reqNamed.add_argument(
        "--out-file",
        action="store",
        dest="out_file",
        help="Output path for .dat file",
        required=True,
    )
    parser_parsingBG_reqNamed.add_argument(
        "--genome-dir",
        action="store",
        dest="genome_dir",
        help="Directory where fasta files are stored",
    )
    # Optional args for the parsingBG command
    parser_parsingBG.add_argument(
        "--bck-var",
        action="store_false",
        default=True,
        dest="only_coverage",
        help="Parse variants for background reads",
    )

    # Create the parser for the parsingCLIP command
    parser_parsingCLIP = subparsers.add_parser(
        "parsingCLIP",
        help="parsingCLIP help",
        description="Parsing the CLIP files.",
        parents=[parent_parsing],
    )
    parser_parsingCLIP_reqNamed = parser_parsingCLIP.add_argument_group(
        "required arguments"
    )
    parser_parsingCLIP_reqNamed.add_argument(
        "--clip-files",
        action="append",
        dest="clip_libs",
        help="BAM files for CLIP libraries",
        required=True,
    )
    parser_parsingCLIP_reqNamed.add_argument(
        "--db-file",
        action="store",
        dest="db_file",
        help="Path to the .GFF.DB file",
        required=True,
    )
    parser_parsingCLIP_reqNamed.add_argument(
        "--out-file",
        action="store",
        dest="out_file",
        help="Output path for .dat file",
        required=True,
    )
    parser_parsingCLIP_reqNamed.add_argument(
        "--genome-dir",
        action="store",
        dest="genome_dir",
        help="Directory where fasta files are stored",
    )
    # Optional args for the parsingCLIP command
    parser_parsingCLIP.add_argument(
        "--mask-miRNA",
        action="store_true",
        dest="mask_miRNA",
        help="Mask miRNA positions",
        default=False,
    )
    parser_parsingCLIP.add_argument(
        "--mask-ovrlp",
        action="store_true",
        dest="mask_ovrlp",
        help="Ignore overlapping gene regions for diagnostic event model fitting",
        default=False,
    )
    parser_parsingCLIP.add_argument(
        "--min-coverage",
        action="store",
        dest="min_coverage",
        help="Minimum summed coverage on a whole gene to be considered in the analysis",
        default=100,
    )
    parser_parsingCLIP.add_argument(
        "--min-peak",
        action="store",
        dest="min_peak",
        help="Minimum maximum peak heigth on a gene to be considered in the analysis",
        default=5,
    )

    # Create the parser for the run_omniCLIP command
    parser_run_omniCLIP = subparsers.add_parser(
        "run_omniCLIP",
        help="run_omniCLIP help",
        description="running the main omniCLIP program.",
    )
    parser_run_omniCLIP_reqNamed = parser_run_omniCLIP.add_argument_group(
        "required arguments"
    )
    parser_run_omniCLIP_reqNamed.add_argument(
        "--bg-dat",
        action="store",
        dest="dat_file_bg",
        help="Path to the parsed background .dat file",
        required=True,
    )
    parser_run_omniCLIP_reqNamed.add_argument(
        "--clip-dat",
        action="store",
        dest="dat_file_clip",
        help="Path to the parsed CLIP .dat file",
        required=True,
    )
    parser_run_omniCLIP_reqNamed.add_argument(
        "--out-dir",
        action="store",
        dest="out_dir",
        help="Output directory for results",
    )
    parser_run_omniCLIP_reqNamed.add_argument(
        "--db-file",
        action="store",
        dest="gene_anno_file",
        help="File where gene annotation is stored",
    )
    # Optional args for the run_omniCLIP command
    parser_run_omniCLIP.add_argument(
        "--bg-type",
        action="store",
        dest="bg_type",
        help="Background type",
        choices=["None", "Coverage", "Coverage_bck"],
        default="Coverage_bck",
    )
    parser_run_omniCLIP.add_argument(
        "--max-it",
        action="store",
        dest="max_it",
        help="Maximal number of iterations",
        type=int,
        default=20,
    )
    parser_run_omniCLIP.add_argument(
        "--tol-log-like",
        action="store",
        dest="tol_lg_lik",
        help="tolerance for lok-likelihood",
        type=float,
        default=10000.0,
    )
    parser_run_omniCLIP.add_argument(
        "--nr_mix_comp",
        action="store",
        dest="nr_mix_comp",
        help="Number of diagnostic events mixture components",
        type=int,
        default=1,
    )
    parser_run_omniCLIP.add_argument(
        "--ign-diag",
        action="store_true",
        dest="ign_diag",
        help="Ignore diagnostic event model for scoring",
        default=False,
    )
    parser_run_omniCLIP.add_argument(
        "--ign-GLM",
        action="store_true",
        dest="ign_GLM",
        help="Ignore GLM model for scoring",
        default=False,
    )
    parser_run_omniCLIP.add_argument(
        "--snp-ratio",
        action="store",
        dest="snps_thresh",
        help="Ratio of reads showing the SNP",
        type=float,
        default=0.2,
    )
    parser_run_omniCLIP.add_argument(
        "--snp-abs-cov",
        action="store",
        dest="snps_min_cov",
        help="Absolute number of reads covering the SNP position",
        type=float,
        default=10,
    )
    parser_run_omniCLIP.add_argument(
        "--norm_class",
        action="store_true",
        dest="norm_class",
        help="Normalize class weights during glm fit",
        default=False,
    )
    parser_run_omniCLIP.add_argument(
        "--seed",
        action="store",
        dest="rnd_seed",
        help="Set a seed for the random number generators",
        default=None,
    )
    parser_run_omniCLIP.add_argument(
        "--diag_event_mod",
        action="store",
        dest="diag_event_mod",
        help="Diagnostic event model",
        choices=["DirchMult", "DirchMultK"],
        default="DirchMultK",
    )
    parser_run_omniCLIP.add_argument(
        "--glm_weight",
        action="store",
        dest="glm_weight",
        help="weight of the glm score with respect to the diagnostic event score",
        type=float,
        default=-1.0,
    )
    parser_run_omniCLIP.add_argument(
        "--skip_diag_event_mdl",
        action="store_true",
        dest="skip_diag_event_mdl",
        help="Do not model the diagnostic events",
        default=False,
    )
    parser_run_omniCLIP.add_argument(
        "--pv",
        action="store",
        dest="pv_cutoff",
        help="bonferroni corrected p-value cutoff for peaks in bed-file",
        type=float,
        default=0.05,
    )
    parser_run_omniCLIP.add_argument(
        "--emp-var",
        action="store_true",
        dest="emp_var",
        help="use the empirical variance if it larger than the expected variance",
        default=False,
    )
    parser_run_omniCLIP.add_argument(
        "--diag-bg",
        action="store_true",
        dest="diag_bg",
        help="estimate diagnostic events for the background states on the background",
        default=False,
    )
    parser_run_omniCLIP.add_argument(
        "--fg_pen",
        action="store",
        dest="fg_pen",
        help="Penalty for fg during scoring",
        type=float,
        default=0.0,
    )
    parser_run_omniCLIP.add_argument(
        "--filter-snps",
        action="store_true",
        dest="filter_snps",
        help="Do not fit diagnostic events at SNP-positions",
        default=False,
    )
    parser_run_omniCLIP.add_argument(
        "--no-subsample",
        action="store_false",
        default=True,
        dest="subs",
        help="Disable subsampling for parameter estimations (Warning: Leads to slow estimation)",
    )
    parser_run_omniCLIP.add_argument(
        "--ign_out_rds",
        action="store_true",
        dest="ign_out_rds",
        help="ignore reads where the ends map ouside of the genes",
        default=False,
    )
    # Runtime args for the run_omniCLIP command (using a different subparser?)
    parser_run_omniCLIP.add_argument(
        "--nb-cores",
        action="store",
        dest="nb_proc",
        help="Number of cores o use",
        type=int,
        default=1,
    )
    parser_run_omniCLIP.add_argument(
        "--save-tmp",
        action="store_true",
        dest="safe_tmp",
        help="Safe temporary results",
        default=False,
    )
    parser_run_omniCLIP.add_argument(
        "--tmp-dir",
        action="store",
        dest="tmp_dir",
        help="Output directory for temporary results",
        default=None,
    )
    parser_run_omniCLIP.add_argument(
        "--verbosity",
        action="store",
        dest="verbosity",
        help="verbosity: 0 (default) gives information of current state of site prediction, 1 gives aditional output on runtime and meomry consupmtiona and 2 shows selected internal variables",
        type=int,
        default=0,
    )
    parser_run_omniCLIP.add_argument(
        "--mask-ovrlp",
        action="store_true",
        dest="mask_ovrlp",
        help="Ignore overlapping gene regions for diagnostic event model fitting",
        default=False,
    )

    # Parsing the arguments if only sites should be predicted
    args = parser.parse_args()

    if args.command == "generateDB":
        generateDB(args)
    elif args.command == "parsingBG":
        parsingBG(args)
    elif args.command == "parsingCLIP":
        parsingCLIP(args)
    elif args.command == "run_omniCLIP":
        run_omniCLIP(args)


if __name__ == "__main__":
    main()
